require "defines"
require "utils"

remote.addinterface("sspl",
{
	clean = function()
		glob.sspl = {	splitSets = {},
						splitSetsToRemove = {},
						version = utils.currentVersion
				   }
	end
})

game.oninit(function() OnInit() end)
game.onload(function() OnLoad() end)
game.onevent(defines.events.ontick, function(event) OnTick(event) end)
game.onevent(defines.events.onbuiltentity, function(event) OnBuiltEntity(event) end)
game.onevent(defines.events.onentitydied, function(event) OnEntityDied(event) end)
game.onevent(defines.events.onpreplayermineditem, function(event) OnPrePlayerMinedItem(event) end)


function OnInit()
	--game.player.print("RailSwitch: OnInit")
	if glob.sspl == nil then
		glob.sspl = {	splitSets = {},
						splitSetsToRemove = {},
						version = utils.currentVersion
				   }
	end
	if not glob.sspl.version then
		game.player.print("Migrating SmartSplitters from unversioned to version " .. utils.currentVersion)
		utils.migrations.versionless()
		game.player.print("Migration successful.")
	end
	if glob.sspl.version ~= utils.currentVersion then
		game.player.print("Migrating SmartSplitters from version " .. glob.sspl.version .. " to version " .. utils.currentVersion .. ".")
		utils.migrations[glob.sspl.version]()
		game.player.print("Migration successful.")
	end
end

function OnLoad()
	--game.player.print("TeleBelts: OnLoad")
	-- Call OnInit to ensure glob.teles exists.
	OnInit()
end

function OnTick(_Event)
	-- Check if there's any sets with invalid splitters and remove those from the splitter lists.
	-- Then check if there's any sets with zero splitters and remove those from the global set table.
	if game.tick % 60 == 0 then
		local toRemove = {}
		for setID, set in pairs(glob.sspl.splitSets) do
			splitterToRemove = {}
			for splitterID, splitter in pairs(set.splitters) do
				if not splitter.valid then
					table.insert(splitterToRemove, splitterID)
				end
			end
			for _, splitterID in pairs(splitterToRemove) do
				table.remove(set.splitters, splitterID)
			end
			if #set.splitters < 1 then
				table.insert(toRemove, setID)
			end
		end
		for _, setID in pairs(toRemove) do
			table.remove(glob.sspl.splitSets)
		end
	end
	
	-- Check for new filters 6 times per second.
	if game.tick % 10 == 0 then
		for _, set in pairs(glob.sspl.splitSets) do
			RecalculateSet(set)
		end
	end
	
	for _, set in pairs(glob.sspl.splitSets) do
		for _, splitter in pairs(set.splitters) do
			if splitter.valid then
				local scanArea = GetScanArea(splitter.direction, splitter.position)
			
				local scan = {	left = game.findentitiesfiltered{	name = "item-on-ground",
										  				 			area = scanArea.left
																},
								right = game.findentitiesfiltered{	name = "item-on-ground",
											  					 	area = scanArea.right
																 }
				 			 }
				
				for sideName, side in pairs(scan) do
					for _, item in pairs(side) do
						local targetSplitterSet = nil
						if set.filteredItems[item.stack.name] then
							targetSplitterSet = set.filteredItems[item.stack.name]
						else
							targetSplitterSet = set.unfilteringSplitters
						end
						
						if not set.itemCounts[item.stack.name] then
							set.itemCounts[item.stack.name] = 0
						else
							set.itemCounts[item.stack.name] = set.itemCounts[item.stack.name] + 1
						end
						
						local moved = false
						local tries = 0
						while not moved and tries < #targetSplitterSet do
							local targetSplitter = targetSplitterSet[((set.itemCounts[item.stack.name] + tries) % #targetSplitterSet) + 1]
							
							moved = MoveStack(item, targetSplitter, sideName)
							tries = tries + 1
						end
					end
				end
			end
		end
	end
end

local directionToOffset = {[0] = {x = -1, y =  0},
						   [2] = {x =  0, y =  1},
						   [4] = {x =  1, y =  0},
						   [6] = {x =  0, y = -1}}
						   
function OnBuiltEntity(_Event)
	local _Entity = _Event.createdentity
	if IsSmartSplitter(_Entity) then
		local _Position = _Entity.position
		local _Direction = _Entity.direction
		
		-- Check if there's a splitter to the left.
		local scanPosl = {x = _Position.x + directionToOffset[_Direction].x, y = _Position.y + directionToOffset[_Direction].y}
		local scanAreal = {{scanPosl.x - 0.25, scanPosl.y - 0.25}, {scanPosl.x + 0.25, scanPosl.y + 0.25}}
		local scanl = game.findentitiesfiltered{name = "smartsplitter", area = scanAreal}
		
		-- Check if there's a splitter to the right.
		local scanPosr = {x = _Position.x - directionToOffset[_Direction].x, y = _Position.y - directionToOffset[_Direction].y}
		local scanArear = {{scanPosr.x - 0.25, scanPosr.y - 0.25}, {scanPosr.x + 0.25, scanPosr.y + 0.25}}
		local scanr = game.findentitiesfiltered{name = "smartsplitter", area = scanArear}
		
		
		local lset = nil
		local rset = nil
		
		local lsetID = nil
		if #scanl == 1 and scanl[1].direction == _Direction then
			lsetID = GetAppropriateSetID(scanl[1])
		end
		local rsetID = nil
		if #scanr == 1 and scanr[1].direction == _Direction then
			rsetID = GetAppropriateSetID(scanr[1])
		end
		
		local set = CreateNewSet()
		if lsetID and rsetID then
			for _, v in pairs(glob.sspl.splitSets[lsetID].splitters) do
				table.insert(set.splitters, v)
			end
			table.insert(set.splitters, _Entity)
			for _, v in pairs(glob.sspl.splitSets[rsetID].splitters) do
				table.insert(set.splitters, v)
			end
			
			table.remove(glob.sspl.splitSets, lsetID)
			table.remove(glob.sspl.splitSets, rsetID)
			--game.player.print("Merged sets.")
		elseif lsetID then
			--game.player.print("Added to left set.")
			set = glob.sspl.splitSets[lsetID]
			table.insert(set.splitters, _Entity)
			table.remove(glob.sspl.splitSets, lsetID)
		elseif lsetID then
			--game.player.print("Added to right set.")
			set = glob.sspl.splitSets[rsetID]
			table.insert(set.splitters, _Entity, 1)
			table.remove(glob.sspl.splitSets, rsetID)
		else
			--game.player.print("Created new set.")
			table.insert(set.splitters, _Entity)
		end
		
		table.insert(glob.sspl.splitSets, set)
		
		RecalculateSet(set)
	end
end

function OnEntityDied(_Event)
	if IsSmartSplitter(_Event.entity) then
		RemoveSmartSplitter(_Event.entity)
	end
end

function OnPrePlayerMinedItem(_Event)
	if IsSmartSplitter(_Event.entity) then
		RemoveSmartSplitter(_Event.entity)
	end
end

function IsSmartSplitter(_Entity)
	return (_Entity.name == "smartsplitter")
end

function CreateNewSet()
	return {	splitters = {},
				filteredItems = {},
				itemCounts = {},
				filteringSplitters = {},
				unfilteringSplitters = {}
		   }
end

function RemoveSmartSplitter(_Splitter)
	local setID, splitterI = GetAppropriateSetID(_Splitter)
	
	if setID then
		local set = glob.sspl.splitSets[setID]
		-- If the splitter is at the edge.
		if splitterI == 1 or splitterI == #set.splitters then
			table.remove(set.splitters, splitterI)
			RecalculateSet(set)
		else
			-- Splitter is somewhere in the middle, split the set.
			leftSet = CreateNewSet()
			rightSet = CreateNewSet()
			
			for i = 1, splitterI - 1 do
				table.insert(leftSet.splitters, set.splitters[i])
			end
			for i = splitterI + 1, #set.splitters do
				table.insert(rightSet.splitters, set.splitters[i])
			end
			
			table.remove(glob.sspl.splitSets, setID)
			
			table.insert(glob.sspl.splitSets, leftSet)
			RecalculateSet(leftSet)
			
			table.insert(glob.sspl.splitSets, rightSet)
			RecalculateSet(rightSet)
		end
	end
end

function RemoveSet(_Set)
	
end

function RecalculateSet(_Set)
	_Set.unfilteringSplitters = {}
	_Set.filteredItems = {}
	_Set.filteringSplitters = {}
	for splitterID, splitter in pairs(_Set.splitters) do
		if splitter.valid then
			-- Fill item filter table.
			for i = 1, 5 do
				local itemName = splitter.getfilter(i)
				if itemName then
					table.insert(_Set.filteringSplitters, splitterID, splitter)
					if _Set.filteredItems[itemName] then
						table.insert(_Set.filteredItems[itemName], splitter)
					else
						_Set.filteredItems[itemName] = {[1] = splitter}
					end
				end
			end
			if not _Set.filteringSplitters[splitterID] then
				table.insert(_Set.unfilteringSplitters, splitter)
			end
		end
	end
end

function GetAppropriateSetID(_Entity)
	for setID, set in pairs(glob.sspl.splitSets) do
		for splitterI, splitter in pairs(set.splitters) do
			if splitter.equals(_Entity) then
				return setID, splitterI
			end
		end
	end
	game.player.print("ERROR: No set ID found for smartsplitter at [" .. _Entity.position.x .. ", " .. _Entity.position.y .. "]")
	return nil, nil
end

function GetScanArea(_Direction, _Position)
	local offL = 0.5
	local offH = 0.3
	if _Direction == 0 then
		return {left  = {{_Position.x + -offH, _Position.y        }, {_Position.x       , _Position.y + offL}},
				right = {{_Position.x        , _Position.y        }, {_Position.x + offH, _Position.y + offL}}
			   }
	elseif _Direction == 2 then
		return {left  = {{_Position.x + -offL, _Position.y + -offH}, {_Position.x       , _Position.y       }},
				right = {{_Position.x + -offL, _Position.y        }, {_Position.x       , _Position.y + offH}}
			   }
	elseif _Direction == 4 then
		return {left  = {{_Position.x        , _Position.y + -offL}, {_Position.x + offH, _Position.y       }},
				right = {{_Position.x + -offH, _Position.y + -offL}, {_Position.x       , _Position.y       }}
			   }
	else
		return {left  = {{_Position.x        , _Position.y        }, {_Position.x + offL, _Position.y + offH}},
				right = {{_Position.x        , _Position.y + -offH}, {_Position.x + offL, _Position.y       }}
			   }
	end
end

local VV = 0.55
local targetOff = {[0] = {left = {x = -0.23, y = -VV  }, right = {x =  0.23, y = -VV  }},
				   [2] = {left = {x =  VV  , y = -0.23}, right = {x =  VV  , y =  0.23}},
				   [4] = {left = {x =  0.23, y =  VV  }, right = {x = -0.23, y =  VV  }},
				   [6] = {left = {x = -VV  , y =  0.23}, right = {x = -VV  , y = -0.23}}
						 }

function MoveStack(_Item, _Splitter, _Side)
	local newPos = {x = _Splitter.position.x + targetOff[_Splitter.direction][_Side].x,
					y = _Splitter.position.y + targetOff[_Splitter.direction][_Side].y}

	if _Splitter.energy < 1 then
		return false
	end


	-- Check if the target location is empty.
	local possiblePos = game.findnoncollidingposition("item-on-ground", newPos, 0.01, 0.01)
	if possiblePos then
		game.createentity{position=possiblePos, name="item-on-ground", stack={name=_Item.stack.name, count=_Item.stack.count}}
		_Item.destroy()

		return true
	end
	return false
end